/*
		JAS v1.00							  (c) 1996 NikSoft International
		--------------------------------------------------------------------
		INPUTM.C

		Input buffer management and error reporting function.

		Monday, November 10, 1997	Version 1.00
*/

#include <stdarg.h>
#include <unistd.h>

#include "globals.h"
#include "jaslib.h"
#include "errmsgs.h"
#include "lexan.h"
#include "symmngr.h"

#define POOL_SIZE					1024
#define BUF_SIZE					1024
#define CURRENT_BUFFER				scanning_buffer
#define STACK_SIZE					20
#define EOB(x)						\
	(&(x)->mem_buffer[x->size] <= (x)->curr_char)

struct buffer_t {
	FILE	*src_file;
	char	*mem_buffer;
	char	*curr_char;
	int		size;
	int		max_size;
};

typedef struct buffer_t *SCAN_BUFFER;

struct context_t {
	int     line_no;
	char	*name;
	char    **Param;
	int     Times2Repeat;
	int     Counter;
	SCAN_BUFFER     buffer;
	SYMBOL          *ContextSym;
};

static SCAN_BUFFER	create_buffer(FILE *, int);
static SCAN_BUFFER	scan_string(char *);
static void			delete_buffer(SCAN_BUFFER);
static int			fill_buffer(SCAN_BUFFER);
static int			rewind_buffer(SCAN_BUFFER);
static int			read_char();

static int pushContext();
static int popContext();

static int ExpandParameter(char *);

/* ---- Local datas -------------------------------------------- */

static struct context_t ContextStack[STACK_SIZE];
static struct context_t *ContextStackPtr= NULL;

static char			*yyname= NULL;
static SYMBOL		*CurrMacroSym;
static char			**ParamBuffers;
static SCAN_BUFFER	scanning_buffer;

static int		RepeatCounter= 1;
static int		RepeatTimes  = 1;

static unsigned Flags= EXPAND;	/* Scan flags. */
static char		scratch[512];	/* Temporary shared buffer. */

extern char ParamPool[];
extern int	PoolPtr;

/* -------- Public functions ------------------------------------------ */


/*
	Function......: SetScanFlag
	Description...:
	Parameters....: None. */
unsigned SetScanFlag(unsigned newFlags)
{
	unsigned oldFlags= Flags;

	Flags |= newFlags;
	return oldFlags;
}


/*
	Function......: UnsetScanFlag
	Description...:
	Parameters....: None. */
unsigned UnsetScanFlag(unsigned newFlags)
{
	unsigned oldFlags= Flags;

	Flags &= (~newFlags);
	return oldFlags;
}


/*
	Function......: GetMacroSym
	Description...:
	Parameters....: None. */
SYMBOL *GetMacroSym()
{
	return CurrMacroSym;
}


/*
	Function......: eat_white_space
	Description...: Eats white spaces.
					Returns the first non-white char on the stream.
	Parameters....: None. */
int eat_white_space()
{
	register int c;

	do {
	    c= read_char();
	} while ( c == ' ' || c == '\t' || c == '\b' );
	unput(c);

	return c;
}


/*
	Function......: input
	Description...: Gets a new char from the current scanning buffer.
					The function performs macro parameters expansion
					and scanning buffer switching.
	Parameters....: None. */
int input()
{
	register int c, c2;

	/* Get a char from the current buffer. */
	c= read_char();

	/* See if it may be a macro parameter label. */
	if ( c != '&' || !(Flags & EXPAND)  ) return c;

	/* Read another char from the buffer. */
	c2= read_char();

	/* If this is not an identifier we haven't found a label, so
	   push back the additional char and return the first we have read. */
	if ( !idechar(c2) ) {
		unput(c2);
		return c;
	};

	/* Read the parameter label */
	c= 0;
	while ( idechar(c2) || isdigit(c2) ) {
		scratch[c++]= (char)c2;
		c2= read_char();
	};

	/* Push back the last char we have read. */
	unput(c2);

	/* NULL terminate the string containing the IDE. */
	scratch[c]= '\0';

	/* Now try to expand the parameter... */
	if ( ExpandParameter(scratch) == -1 ) {
		if ( !(Flags & IGNORE) ) error(ERR_NONFATAL, INVPARAM, scratch);
		return '&';
	};

	return input();
}


/*
	Function......: unput
	Description...: Puts a character back into the current scanning
					buffer, subsequent read from the buffer will return
					this char.
					Returns the passed parameter if successful,
					-1 otherwise.
	Parameters....:
					c		character to be put back into the current
							buffer. */
int unput(int c)
{
	long filepos;
	long newpos;

	if ( CURRENT_BUFFER->curr_char == CURRENT_BUFFER->mem_buffer ) {
		if ( !CURRENT_BUFFER->src_file ) return -1;

		/* Move back file pointer */
		filepos= ftell(CURRENT_BUFFER->src_file);
		newpos= -(CURRENT_BUFFER->size + CURRENT_BUFFER->max_size);
		if ( fseek(CURRENT_BUFFER->src_file, newpos, SEEK_CUR) == -1 ) {
			fseek(CURRENT_BUFFER->src_file, filepos, SEEK_SET);
			return -1;
		};

		/* Now refill the current buffer */
		fill_buffer(CURRENT_BUFFER);

		/* Point to the last character of the buffer. */
		CURRENT_BUFFER->curr_char=
			CURRENT_BUFFER->mem_buffer + CURRENT_BUFFER->size;
	};

	*--(CURRENT_BUFFER->curr_char)= (char)c;
	return c;
}


/*
	Function......: error
	Description...: Signals an error taking proper action depending on
				    the flags of the error.
	Parameters....:
				    ErrFlags		Error flags
					ErrNum			Error Number (index for the table of
									the error messages)
					...				Parameters needed by the particular
									error number */
void error(int ErrFlags, int ErrNum, ...)
{
	va_list arg_ptr;
	static char *ErrType[]= {
		"Warning", "Error", "Fatal", "Panic"
	};

	va_start(arg_ptr, ErrNum);

	if ( !(ErrFlags & ERR_NOFILE) ) {
		fprintf(stderr, "%s:", yyname);
		if ( CurrMacroSym != NULL ) {
			fprintf(stderr, "%lu(%s:%u): ", CurrMacroSym->Value,
				CurrMacroSym->Name, yylineno);
		} else {
			fprintf(stderr, "%u: ", yylineno);
		}
	};

	/* Print the error type */
	fprintf(stderr, "%s: ", ErrType[ErrFlags & ERR_MASK]);

	/* Print the error message */
	if ( ErrNum >= MAXERROR ) ErrNum = MAXERROR;

	vfprintf(stderr, ErrMsgs[ErrNum], arg_ptr);
	va_end(arg_ptr);

	/* If an error is detected, update error counter. */
	if ( ( ErrFlags & ERR_MASK ) == 1 ) {
		Errors++;
	};

	switch ( (ErrFlags & ERR_MASK) ) {
		case ERR_NONFATAL:
	        if ( !--MaxErrors ) {
	            /* Too many errors, aborting ... */
		        fputs("Too many error detected.", stderr);

		case ERR_FATAL:
		case ERR_PANIC:
		        fprintf(stderr, "Exiting and removing %s.\n",
					getOutputFileName() );
		        unlink( getOutputFileName() );
		        exit(-1);
	        };
	};

	/* Print the usage message if requested */
	if ( ErrFlags & ERR_USAGE ) fputs(UsageMsg, stderr);
}


/*
	Function......: SwitchToFile
	Description...: Switch the scanner input buffer to the file passed
					as parameter. The current input file is put in the
					context stack if this one is not full.
					Return 0 if successfull, -1 otherwise.
	Parameters....:
					filename			name of the file to switch to.*/
int SwitchToFile(char *filename)
{
	register FILE *NewIn;

	if ( !filename ) return -1;

	/* NOTE: if 'ContextStackPtr' is equal to NULL this is the first file
	   the scanner is switching to, i.e. the source file to be compiled
	   so we must use the default error reporting behaviour of the
	   library functions */
	if ( ContextStackPtr ) LIB_setflags(ERR_NONFATAL);
	NewIn= LIB_fopen(filename, M_RDTEXT);

	/* Restore the default error reporting flags and save the
	   current context on the stack. */
	if ( ContextStackPtr ) {
		LIB_setflags( -1 );
		pushContext();
	} else {
		/* Initialize the ContextStack pointer. */
		ContextStackPtr= (ContextStack - 1);
	}

	/* Switch to the new file context. */
	if ( !NewIn ) return -1;

	yylineno      = 1;
	yyname        = filename;
	CURRENT_BUFFER= create_buffer(NewIn, BUF_SIZE);

	return 0;
}


/*
	Function......: SwitchToMacro
	Description...: Switches the lexical context, i.e. the buffer from
					which the scanning is performed, to the one pointed
					by the identifier allowing a sort of macro processing.
	Parameters....:
					MacroSym		pointer to the buffer.
					ParamsText		pointer to the parameters string. */
int SwitchToMacro(SYMBOL *MacroSym, char *ParamsText)
{
	register char *source = NULL;
	register char **param_ptrs = NULL;
	int count= 0;
	int size = strlen(ParamsText);

	if ( !ParamsText || size == 0 ) {
		param_ptrs= NULL;
		goto no_params;
	} else if ( ( PoolPtr + size ) >= POOL_SIZE ) {
	/* We can't allocate the parameters string in the Pool, we are
	   out of free space. */
		source = LIB_strdup(ParamsText);
	} else {
	/* Allocate the string in the Pool. */
		source  = strcpy(&ParamPool[PoolPtr], ParamsText);
		PoolPtr += size + 1;
	};

	size= 0;
	do {
		if ( size == count ) {
			size += 8;
			param_ptrs= (char **)LIB_realloc(param_ptrs,
				sizeof(char *) * (size + 1));
		};
		param_ptrs[count]= source;

		if ( count++ == MacroSym->ExtraLong ) {
			error(ERR_WARNING, EXTRAPARAMERR, MacroSym->Name);
		};

		while ( *source && *source != ',' ) source++;
		if ( *source ) *source++ = '\0';

		/* Skip white spaces. */
		while ( *source && isspace(*source) ) source++;
	} while ( *source );

	/* NULL terminate the parameters array. */
	param_ptrs[count]= NULL;

no_params:
	pushContext();

	/* Register the current number of line in the Value field */
	MacroSym->Value = yylineno;

	/* Defines the new context */
	CurrMacroSym	= MacroSym;
	ParamBuffers	= param_ptrs;
	RepeatTimes		= 1;
	RepeatCounter	= 1;
	yylineno		= 1;

	/* Switch the lexer to the new buffer */
	scan_string( ((struct macro_t *)MacroSym->ExtraPtr)->MacroText );

	return 0;
}


/*
	Function......: SwitchToText
	Description...: Switches the lexical context to the text_t buffer
					pointed by the parameter.
	Parameters....:
					NewText			pointer to the text_t struct.
					Times			number of repetition for the buffer.*/
int SwitchToText(char *NewText, int Times)
{
	pushContext();

	/* Defines the new context */
	ParamBuffers	= NULL;
	CurrMacroSym	= NULL;
	RepeatTimes		= Times;
	RepeatCounter	= 1;

	/* Switch the lexer to the new buffer */
	scan_string( NewText );

	return 0;
}

/* -------- Static functions ------------------------------------------ */

/*
	Function......: create_buffer
	Description...: Creates a new scanning buffer.
					Returns a pointer to the newly created buffer
					if successful.
	Parameters....: None. */
static SCAN_BUFFER create_buffer(FILE *src, int buff_size)
{
	register SCAN_BUFFER ret_val;

	ret_val= (SCAN_BUFFER)LIB_malloc( sizeof(struct buffer_t) );

	/* Fill in structure fields */
	ret_val->src_file   = src;
	ret_val->curr_char  = ret_val->mem_buffer= (char *)LIB_malloc(buff_size);
	ret_val->size		= 0;
	ret_val->max_size   = buff_size;

	return ret_val;
}


/*
	Function......: delete_buffer
	Description...: Destroys the specified SCAN_BUFFER
	Parameters....: None. */
static void delete_buffer(SCAN_BUFFER ptr)
{
	if ( ptr == NULL ) return;

	if ( ptr->src_file ) fclose(ptr->src_file);
	LIB_free( ptr->mem_buffer );
	LIB_free( ptr );
}


/*
	Function......: fill_buffer
	Description...: Reads up to dest->max_size chars from the source
					file into the dest->mem_buffer memory buffer.
					Returns 0 on success, -1 otherwise.
	Parameters....:
					dest		pointer to the buffer to fill up. */
static int fill_buffer(SCAN_BUFFER dest)
{
	register int bytesread;

	if ( !dest->src_file ) return -1;

	bytesread= fread(dest->mem_buffer, 1, dest->max_size, dest->src_file);
	if ( bytesread <= 0 ) return -1;

	dest->curr_char= dest->mem_buffer;
	dest->size= bytesread;
	return 0;
}


/*
	Function......: scan_string
	Description...:
	Parameters....: None. */
static SCAN_BUFFER scan_string(char *text)
{
	register SCAN_BUFFER ret_val;

	/* Creates a new scanning buffer. */
	ret_val= create_buffer(NULL, strlen(text) + 1);

	/* Fill the memory buffer with the string. */
	strcpy(ret_val->mem_buffer, text);
	ret_val->size= ret_val->max_size - 1;

	/* Switch to the new scanning buffer */
	CURRENT_BUFFER= ret_val;
	return ret_val;
}


/*
	Function......: rewind_buffer
	Description...: Rewinds a scanning buffer.
					Returns 0 on success, -1 otherwise.
	Parameters....:
					dest		Buffer to rewind. */
static int rewind_buffer(SCAN_BUFFER dest)
{
	if ( dest->src_file != NULL ) {
		return fseek(dest->src_file, 0, SEEK_SET);
	};

	dest->curr_char= dest->mem_buffer;
	return 0;
}


/*
	Function......: read_char
	Description...: Gets and returns a char from the current
					scanning buffer, if not at end of buffer,
					EOF otherwise.
	Parameters....: None. */
static int read_char()
{
	if ( EOB(CURRENT_BUFFER) && fill_buffer(CURRENT_BUFFER) == -1 ) {
		if ( RepeatCounter++ < RepeatTimes ) {
			rewind_buffer(CURRENT_BUFFER);
		} else if ( popContext() == -1 ) {
			return EOF_CHAR;
		} else {
			return read_char();
		};
	};

	return *(CURRENT_BUFFER->curr_char++);
}


/*
	Function......: popContext
	Description...: Pops the current lexical context from the stack.
	Parameters....: None. */
static int popContext()
{
	if ( ContextStackPtr < ContextStack ) return -1;

	/* clean up some unused memory. */
	if ( CurrMacroSym != NULL ) {
		if ( ParamBuffers ) {
			/* De-allocate the memory for the buffers. */
			if ( *ParamBuffers >= ParamPool &&
				 *ParamBuffers < &ParamPool[POOL_SIZE] ) {
			/* The buffer has been allocated in the ParamPool */
				PoolPtr -= (&ParamPool[PoolPtr] - *ParamBuffers);
			} else {
			/* The buffers has been allocated in the heap. */
				LIB_free(*ParamBuffers);
			};

			LIB_free(ParamBuffers);
		};
	    FreeSymbolList(CurrMacroSym);
	};

	/* restore the saved context. */
	yylineno        = ContextStackPtr->line_no;
	yyname			= ContextStackPtr->name;
	CurrMacroSym    = ContextStackPtr->ContextSym;
	RepeatTimes     = ContextStackPtr->Times2Repeat;
	RepeatCounter   = ContextStackPtr->Counter;
	ParamBuffers	= ContextStackPtr->Param;

	delete_buffer(CURRENT_BUFFER);
	CURRENT_BUFFER= ContextStackPtr->buffer;
	ContextStackPtr--;

	return 0;
}


/*
	Function......: pushContext
	Description...: Pushes the current lexical context onto the stack.
	Parameters....: None. */
static int pushContext()
{
	if ( ContextStackPtr == &ContextStack[STACK_SIZE] ) {
		/* Out of context stack space. */
		error(ERR_FATAL, CNTXTERR);
		return -1;
	};

	/* Save the current scanning buffer on top of the context stack. */
	ContextStackPtr++;
	ContextStackPtr->line_no     = yylineno;
	ContextStackPtr->name		 = yyname;
	ContextStackPtr->ContextSym	 = CurrMacroSym;
	ContextStackPtr->buffer      = CURRENT_BUFFER;
	ContextStackPtr->Times2Repeat= RepeatTimes;
	ContextStackPtr->Counter	 = RepeatCounter;
	ContextStackPtr->Param		 = ParamBuffers;

	return 0;
}


/*
	Function......: ExpandParameter
	Description...: Expands the macro parameter switching to its text
					buffer.
					Return 0 if successfull, -1 otherwise.
	Parameters....:
					ParamName		name of the parameter to expand. */
static int ExpandParameter(char *ParamName)
{
	register char **BuffCrsr= ParamBuffers;
	register char **NameCrsr;
	static char dump[20];
	char *crsr;

	if ( !LIB_stricmp(ParamName, "COUNT") ) {
		sprintf(dump, "%u", RepeatCounter);
		SwitchToText(dump, 1);
		return 0;
	}

	if ( CurrMacroSym != NULL && BuffCrsr ) {
	    NameCrsr= ((struct macro_t *)CurrMacroSym->ExtraPtr)->Parameters;

	    while ( *BuffCrsr && *NameCrsr ) {
		    if ( !LIB_stricmp(ParamName, *NameCrsr) ) {
			    SwitchToText(*BuffCrsr, 1);
			    return 0;
		    }

		    /* Next parameter. */
		    BuffCrsr++;
		    NameCrsr++;
	    }
	};

	if ( !(Flags & IGNORE) ) return -1;

	/* Ignore the parameter,
	   Push back all the chars we have read. */
	for (crsr= ParamName + strlen(ParamName) - 1; crsr >= ParamName; crsr--) {
		unput(*crsr);
	};

	return -1;
}

/* -------- End Of File ----------------------------------------------- */
