/*
		JAS v1.00							  (c) 1996 NikSoft International
		--------------------------------------------------------------------
		JASLIB.C

		Library functions for JAS.
		 - File handling
		 - Memory allocation
		 - String handling

		In general all the public functions of this module are simply
		wrapper around the equivalent standard library functions.
		Code readapted from the original NASM code.

		Saturday, November 30, 1996 Version 1.00
*/

#include <ctype.h>

#include "globals.h"
#include "jaslib.h"

#define DEFAULT_FLAGS			(ERR_FATAL | ERR_NOFILE)

/* -------- Local Datas ----------------------------------------------- */

/* Error flags to use in error() function call */
static int errflags= DEFAULT_FLAGS;

/* -------------------------------------------------------------------- */

/*
	Function......: LIB_setflags
	Description...: Set the errflags local variable.
					Return the old value of the variable.
	Parameters....:
					toSet		new value for the errflags variable.
								If this is equal to -1 errflag is reseted
								to the default value.
*/
int LIB_setflags(int toSet)
{
	int old_val= errflags;

	errflags= ( toSet == -1 ? DEFAULT_FLAGS : toSet);
	return old_val;
}

void *LIB_malloc(size_t size)
{
	register void *p = malloc(size);

	if ( !p ) error(errflags, MEMERR);

	/* Clear the allocated area */
	bzero(p, size);

	return p;
}

void *LIB_realloc(void *q, size_t size)
{
	register void *p = q ? realloc(q, size) : malloc(size);

	if ( !p ) error(errflags, MEMERR);

	return p;
}

void LIB_free(void *q)
{
	free(q);
}

char *LIB_strdup(char *s)
{
	char *p;

	p = LIB_malloc( strlen(s) + 1 );
	return strcpy(p, s);
}

char *LIB_strcat(char *s1, char *s2)
{
	register char *RetVal;

	RetVal = (char *)LIB_malloc(strlen(s1) + strlen(s2) + 1);
	strcpy(RetVal, s1);

	return strcat(RetVal, s2);
}

int LIB_stricmp(const char *s1, const char *s2)
{
	while (*s1 && toupper(*s1) == toupper(*s2))	{
		s1++;
		s2++;
	};

	if (!*s1 && !*s2) return 0;
	else if (toupper(*s1) < toupper(*s2)) return -1;
	else return 1;
}

int LIB_strnicmp(const char *s1, const char *s2, int n)
{
	while (n > 0 && *s1 && toupper(*s1) == toupper(*s2)) {
	    s1++;
		s2++;
		n--;
	}

	if ((!*s1 && !*s2) || n==0)	return 0;
	else if (toupper(*s1) < toupper(*s2)) return -1;
	else return 1;
}


/*
	Function......: StrToUpper
	Description...: Converts a string to upper case.
				    Returns the converted string.
	Parameters....:
					Src			String that has to be converted.
					Dest		Buffer where the converted string should
								be stored.
*/
char *LIB_StrToUpper(char *Src, char *Dest)
{
	register char *Crsr= Dest;

	while ( (*Crsr++ = toupper(*Src++)) != 0 );

	return Dest;
}


void strip_newline (char *s) {
	if (s[strlen(s)-1] == '\n')
		s[strlen(s)-1] = '\0';
}

#define numvalue(c)					\
	((c)>='a' ? (c)-'a'+10 : (c)>='A' ? (c)-'A'+10 : (c)-'0')


/*
	Function......: LIB_readnum
	Description...: Reads a number from a string, support different format.
					Returns the number read if successfull, return -1 and
					calls error() function if an error is encountered.
					The function doesn't care about overflow, if constant
					exceedes INT_MAX limit the result wraps around.
					NOTE: This function should always be successfull,
						  because is called only by the Lexical Analyzer.
	Parameters....:
					str				Pointer to the source string.
*/
int LIB_readnum(char *str)
{
	register char *r = str;
	register char *q = r + strlen(str);
	int radix;
	int result;

	/* Skip the sign if present */
	if ( *r == '-' || *r == '+' ) r++;

	/* If it begins 0x, 0X or $, or ends in H, it's in hex. if it ends
	   in Q, it's octal. if it ends in B, it's binary.
	   Otherwise, it's ordinary decimal. */
	if ( *r == '0' &&  (r[1]=='x' || r[1]=='X') ) {
		radix = 16;
		r += 2;
	} else if ( *r == '$' ) {
		radix = 16;
		r++;
	} else if ( q[-1] == 'H' || q[-1] == 'h' ) {
		radix = 16;
		q--;
	} else if ( q[-1] == 'O' || q[-1] == 'o' ) {
		radix =	 8;
		q--;
	} else if ( q[-1] == 'B' || q[-1] == 'b' ) {
		radix =	 2;
		q--;
	} else {
		radix = 10;
	}

	result = 0;

	while (*r && r < q) {
	    result = radix * result + numvalue(*r);
	    r++;
	};

	if ( *str == '-' ) result -= 0;

	return result;
}

static long next_seg;

void seg_init(void) {
	next_seg = 0;
}

long seg_alloc(void) {
	return (next_seg += 2) - 2;
}

void fwriteshort(short data, FILE *fp)
{
	fputc(data & 255, fp);
	fputc((data >> 8) & 255, fp);
}

void fwritelong(long data, FILE *fp)
{
	fputc(data & 255, fp);
	fputc((data >> 8) & 255, fp);
	fputc((data >> 16) & 255, fp);
	fputc((data >> 24) & 255, fp);
}


/*
	Function......: LIB_addExtension
	Description...: Adds a standard extension to a file name.
					Returns the processed file name. (outname)
	Parameters....:
					inname				Source name.
					outname				Output name.
					extension			Extension to add.
*/
char *LIB_addExtension(char *inname, char *outname, char *extension)
{
	register char *p, *q;

	q = inname;
	p = outname;

	/* copy, and find end of string */
	while ( *q ) *p++ = *q++;

	/* terminate it */
	*p = '\0';

	/* find final period (or whatever) */
	while (p > outname && *--p != '.');

	if (*p != '.') {
		/* go back to end if none found */
		while (*p) p++;
	}
	else {
		/* File names has already an extension, so don't add the default
		   and return the filename */
		return outname;
	}

	/* is the extension a valid one? */
	if ( !extension || *extension != '.' ) {
		strcpy(p, ".out");
	}
	else {
		strncpy(p, extension, 4);
	};

	return outname;
}


/*
	Function......: LIB_fopen
	Description...: Opens a file using the standard library function.
					Returns a pointer to the opened file FILE struct,
					calls error() and returns NULL otherwise.
	Parameters....:
					filename			pointer to the file to be opened.
					mode				file open mode.
*/
FILE *LIB_fopen(char *filename, int mode)
{
	register FILE *retval;
	register char *fmode;

	switch ( mode ) {
		case M_RDBIN:
			fmode= "rb";
			break;

		case M_RWBIN:
			fmode= "wb";
			break;

		case M_RDTEXT:
			fmode= "rw";
			break;

		case M_RWTEXT:
			fmode= "wb";
			break;

		default:
			return NULL;
	};

	retval= fopen(filename, fmode);

	if ( !retval ) {
		error(errflags, IOERR, filename);
	};

	return retval;
}

/* -------- End Of File ----------------------------------------------- */
