/*
		JAS v1.00							  (c) 1996 NikSoft International
		--------------------------------------------------------------------
		LEXAN.C

		Lexical analyzer.
		Scans the input looking for the following lexemes:
			- numerical constant in different format (hex, oct, bin, dec)
			- string constant
			- identifier
			- keywords (instruction names, special keywords, register names,
			  etc.)

		03/11/96 19.13	Version 1.00

		Tuesday, November 26, 1996
			All the reserved keywords are stored in an hash table
			handled by the symbol manager module. Reserved keywords
			are recognized via the function IsKeyword of the above
			module.
		Tuesday, July 22, 1997
			Added context switching and support for #defined style
			macro.
		Sunday, August 03, 1997
			Improved the context switching capabilities.
			Improved the support to macro and macro processing, now
			.REPEAT macro command is available.
		Thursday, August 07, 1997
			Added the processing for .MACRO directive and macro expansion.
		Monday, November 10, 1997
			The lexical analyzer will be no longer generated by FLEX.
			Due to some limitations during macro parameter expansion
			the module is now written "by hand", the utility and buffer
			switching functions have been moved into the INPUTM.C
			module.
*/

#include <unistd.h>

#include "globals.h"
#include "jaslib.h"

#include "ptypes.h"
#include "symmngr.h"
#include "parser.h"

#include "lexan.h"

#define LEXEME_DELTA			512
#define MAX_LEXEME_LENGTH		1024

#define YYSTATE					yystate
#define BEGIN					yystate=

enum lexical_state {
	INITIAL= 0,
	START_DEF,
	READ_DEF,
	START_MACRO,
	READ_PARAM,
	START_TEXT,
	READ_TEXT,
	READ_FORMAL,
	START_IF,
	READ_IFTEXT,
	READ_ID,
	LAST_STATE
};

extern int TokenInfo;

struct state_t {
	int token;
	int info;
	int pos;
};

/* -------- Function's prototypes ------------------------------------- */

static int	do_IDE(char *);
static int	do_DIRECTIVE(char *);

static int	AddToHeap(char *, int (*)(char *));

static int	macro_body_stop(char *);
static int	if_body_stop(char *);

static void reset();
static int	advance();
static int	next_token();
static int	accept(struct state_t);
static int	read_text(int, int, int);

/* -------- Static Datas ---------------------------------------------- */

static unsigned yysize;

static SYMBOL	*SetMacro;
static int		DefineDepth;

static struct state_t	curr_state;
static struct state_t	accept_state;

static enum lexical_state yystate;
static enum lexical_state old_state;

static int	Instruction= FALSE;
static char *LineHeap;
static int	HeapLines;
static int	HeapSize;
static int	HeapPtr;

/* -------- Public Datas ---------------------------------------------- */

unsigned int yylineno;
unsigned int yylen;
char *yytext;

/* -------- Public functions ------------------------------------------ */

/*
	Function......: yylex
	Description...: Returns the next token from the input source.
	Parameters....: None. */
int yylex()
{
	register int c= 0;

	do {
		reset();

	    /* Eat up white space. */
		c= eat_white_space();

	    switch ( YYSTATE ) {
		    case READ_DEF:
			    read_text('\n', FALSE, FALSE);
			    unput('\n');
			    yylval.strval= LIB_strdup(yytext);
			    BEGIN INITIAL;
			    return TOKEN_STR;

			case READ_PARAM:
				// SetScanFlag(STOP_AT_EOF);
				c= advance();
				if ( idechar(c) ) {
					while ( idechar(c) || isdigit(c) ) {
						c= advance();
					};
					unput(c);
					yytext[--yylen]= '\0';
					yylval.strval= LIB_strdup(yytext);
					return TOKEN_STR;
				} else if ( c == '\n' ) {
					BEGIN READ_TEXT;
					DefineDepth= 0;
				};
				return c;

			case READ_TEXT:
				read_text('\n', TRUE, TRUE);
				if ( !AddToHeap(yytext, macro_body_stop) ) {
				    BEGIN INITIAL;

					yylval.textval.text = LIB_strdup(LineHeap);
					yylval.textval.Lines= HeapLines;

					/* Reset the LineHeap dynamic array */
	                HeapLines= 0;
	                HeapPtr	 = 0;
	                HeapSize = 0;

				    return TOKEN_TEXT;
				};
				yylineno++;
				c= 0;
				break;

			case READ_FORMAL:
				read_text('\n', FALSE, TRUE);
				yylineno++;
				SwitchToMacro(SetMacro, yytext);
				BEGIN INITIAL;
				c= 0;
				break;

			case READ_IFTEXT:
				read_text('\n', TRUE, TRUE);
				if ( !AddToHeap(yytext, if_body_stop) ) {
				    BEGIN INITIAL;

					yylval.textval.text = LIB_strdup(LineHeap);
					yylval.textval.Lines= HeapLines;

					/* Reset the LineHeap dynamic array */
	                HeapLines= 0;
	                HeapPtr	 = 0;
	                HeapSize = 0;

				    return TOKEN_TEXT;
				};
				yylineno++;
				c= 0;
				break;

			case START_TEXT:
				if ( c == '\n' ) {
					BEGIN READ_TEXT;
					DefineDepth= 0;
				    return advance();
				};

			case START_IF:
				if ( c == '\n' ) {
					BEGIN READ_IFTEXT;
					DefineDepth= 0;
				    return advance();
				};

			default:
				if ( YYSTATE >= LAST_STATE ) {
					error(ERR_PANIC, GENERR, "Invalid lexical state!");
				};
				c= next_token();
		};

	} while ( !c );

	/* Return the token. */
	return c;
}


/*
	Function......: SetExpr
	Description...: Set the 'Instruction' flag.
	Parameters....:
					ExprFlag	If non zero tells the lexer that is
								scanning tokens for an expressions,
								i.e. it has already read tokens for a
								directive or instruction. */
void SetExpr(int ExprFlag)
{
	Instruction= (ExprFlag ? TRUE : FALSE);
}


/*
	Function......: BuildMacro
	Description...: Builds up a macro_t struct and ties it to the passed
					identifier. The struct it is build using the text
					currently in the LineHeap array and the passed array of
					parameters. Returns MacroId if successfull, NULL
					otherwise.
	Parameters....:
					MacroId			pointer to the identifier for the
									macro.
					ParamArray		pointer to a NULL terminated array of
									strings, i.e. the names of the
									parameters.
					Body			Pointer to the body of the macro.
					Lines			Number of of the macro body. */
SYMBOL *BuildMacro(SYMBOL *MacroId, char **ParamArray, char *Body, int Lines)
{
	register struct macro_t *dummy;

	dummy = (struct macro_t *)LIB_malloc(sizeof(struct macro_t));

	/* Fill in structure field */
	dummy->MacroText  = Body;
	dummy->lines	  = Lines;
	dummy->Parameters = ParamArray;

	/* Fill up the Extra information fields of the SYMBOL struct. */
	MacroId->ExtraPtr= dummy;

	return MacroId;
}


/*
	Function......: RepeatText
	Description...: Repeats the collected line of text n times.
					This is a orrible implementation of the REPEAT macro
					command.
	Parameters....: Obvious.*/
void RepeatText(char *text, int lines, int n)
{
	SwitchToText(text, n);
	yylineno -= lines;
}

/* -------- Static Functions ------------------------------------------ */


/*
	Function......: next_token
	Description...:
	Parameters....: None. */
static int next_token()
{
	register int c;

	/* Begin the scanning */
	c= advance();

	/* Eat up comment */
	if ( comment(c) ) {
		UnsetScanFlag(EXPAND);
		do {
			c= input();
		} while ( c != '\n' && c != EOF_CHAR );
		SetScanFlag(EXPAND);

		return c;
	};

	for ( ; c != EOF_CHAR; c= advance() ) {
		if ( curr_state.token ) {
			switch ( curr_state.token ) {
				case '.':
			        if ( isdigit(c) ) {
				        curr_state.token= TOKEN_FLT;
						curr_state.info = 1;
			        } else if ( idechar(c) ) {
				        curr_state.token= TOKEN_DIRECTIVE;
			        } else {
				        return accept(accept_state);
			        };
					break;

				case TOKEN_HERE:
			        if ( isxdigit(c) ) {
				        curr_state.token= TOKEN_NUM;
				        curr_state.info = 16;
				        accept_state= curr_state;
			        } if ( idechar(c) ) {
				        curr_state.token= TOKEN_ID;
				        accept_state= curr_state;
			        } else {
				        return accept(accept_state);
			        };
					break;

				case TOKEN_ID:
					/* Process the IDENTIFIER */
					if ( idechar(c) || isdigit(c) ) {
						accept_state= curr_state;
					} else {
						return accept(accept_state);
					};
					break;

				case TOKEN_DIRECTIVE:
					/* Process DIRECTIVE */
					if ( idechar(c) || isdigit(c) ) {
						accept_state= curr_state;
					} else {
						return accept(accept_state);
					};
					break;

				case TOKEN_NUM:
					switch ( curr_state.info ) {
						case 0:
							if ( c == 'x' && *yytext == '0' ) {
								curr_state.info = 16;
								break;
							};
						case 2:
							if ( c == '0' || c == '1' ) {
								curr_state.info= 2;
								accept_state= curr_state;
								break;
							} else if ( c == 'b' || c == 'B' ) {
								accept_state= curr_state;
								curr_state.info = -2;
								c= advance();

								goto last_beach;
							}

						case 8:
							if ( c >= '0' && c <= '7' ) {
								curr_state.info= 8;
								accept_state= curr_state;
								break;
							} else if ( c == 'o' || c == 'O' ) {
								accept_state= curr_state;
								return accept(accept_state);
							};

						case 10:
							if ( isdigit(c) ) {
								curr_state.info= 10;
								accept_state= curr_state;
								break;
							} else if ( c == 'd' || c == 'D' ) {
								accept_state= curr_state;
								curr_state.info = -10;
								c= advance();
							} else if ( c == '.' ) {
								curr_state.token= TOKEN_FLT;
								curr_state.info = 1;
								break;
							};

						case 16:
						last_beach:
						default:
							if ( accept_state.info == 16 || yytext[1] == 'x' ) {
							    if ( !isxdigit(c) ) {
									return accept(accept_state);
								};
								accept_state= curr_state;
							} else if ( (c == 'h' || c == 'H') ) {
								accept_state= curr_state;
								return accept(accept_state);
							} else if ( !isxdigit(c) ) {
								return accept(accept_state);
							};
					};
					break;

				case TOKEN_FLT:
					if ( isdigit(c) ) {
						accept_state= curr_state;
					} else if ( c == '.' && curr_state.info == 0 ) {
						curr_state.info= 1;
						c= advance();
						if ( isdigit(c) ) {
							accept_state= curr_state;
						} else {
							return accept(accept_state);
						};
					} else if ( (c == 'e' || c == 'E') &&
								 curr_state.info < 2 ) {
						curr_state.info= 2;
					} else if ( (c == '-' || c == '+') &&
								 curr_state.info < 3 ) {
						curr_state.info= 3;
					} else {
						return accept(accept_state);
					}
					break;

				default:
					/* Two chars token */
					if ( curr_state.token != c ) {
						return accept(accept_state);
					};

					switch ( c ) {
						case '%':
						case '@':
							c= advance();
							if ( !idechar(c) ) {
								return accept(accept_state);
							};
							curr_state.token= TOKEN_ID;
							curr_state.info = 0;
							accept_state= curr_state;
							break;

						case '>':
							curr_state.token= TOKEN_SHR;
							return accept(curr_state);

						case '<':
							curr_state.token= TOKEN_SHL;
							return accept(curr_state);

					}
				}
		} else if ( c == '.' ) {
			curr_state.token= '.';
			accept_state= curr_state;
		} else if ( c == '$' ) {
			curr_state.token= TOKEN_HERE;
			accept_state= curr_state;
		} else if ( idechar(c) ) {
			curr_state.token= TOKEN_ID;
			accept_state= curr_state;
		} else if ( isdigit(c) ) {
			curr_state.token= TOKEN_NUM;
			curr_state.info = 0;
			accept_state= curr_state;
		} else if ( c == '\\' ) {
			do {
				c= input();
				yylineno += (c == '\n');
			} while ( isspace(c) );
			unput(c);
			reset();
		} else if ( c == '\'' || c == '\"' ) {
			read_text(c, FALSE, FALSE);
			if ( c == '\'' && yylen == 1 ) {
			/* Single char constant */
				yylval.chrval= *yytext;
				return TOKEN_CHR;
			} else {
			/* String constant */
			    yylval.strval= LIB_strdup(yytext);
	            TokenInfo= yylen + (c == '\"');
			    return TOKEN_STR;
			}
		} else {
			/* single char token */
			curr_state.token = c;
			accept_state= curr_state;
		};
	};

	/* Return EOF or the last accepted state. */
	return (accept_state.token ? accept(accept_state) : c);
}


/*
	Function......: reset
	Description...:
	Parameters....: None. */
static void reset()
{
	if ( yytext == NULL ) {
		yytext= LIB_malloc(LEXEME_DELTA);
		yysize= LEXEME_DELTA;
	};

	curr_state.token= 0;
	curr_state.pos	= 0;
	curr_state.info = 0;
	accept_state= curr_state;

	yylen= 0;
	yytext[0]= '\0';
}


/*
	Function......: advance
	Description...:
	Parameters....: None. */
static int advance()
{
	register int c= input();

	if ( curr_state.pos == yysize ) {
		yysize += LEXEME_DELTA;
		yytext= LIB_realloc(yytext, yysize);
	};

	yytext[yylen]= (char)c;
	yylen++;
	curr_state.pos++;

	return c;
}


/*
	Function......: accept
	Description...:
	Parameters....: None. */
static int accept(struct state_t state)
{
	register int i;

	/* Push unrecognized chars back into the input buffer */
	yylen--;
	for (i= yylen; i >= state.pos; i--) {
		unput( yytext[i] );
	};

	/* NULL Terminate the lexeme string */
	yytext[state.pos]= '\0';

	switch ( state.token ) {
		case TOKEN_ID:
			return do_IDE(yytext);

		case TOKEN_DIRECTIVE:
			return do_DIRECTIVE(yytext);

		case TOKEN_NUM:
			yylval.intval= LIB_readnum(yytext);
			return TOKEN_NUM;

		case TOKEN_FLT:
			sscanf(yytext, "%f", &yylval.floatval);
			return TOKEN_FLT;

		case '\n':
			/* Reset the 'Istruction' flag */
			Instruction= FALSE;
	};

	return state.token;
}


/*
	Function......: do_DIRECTIVE
	Description...:
	Parameters....: None. */
static int do_DIRECTIVE(char *name)
{
	register int anInt;

	Instruction= TRUE;

	/* Predefined directive */
	anInt= IsKeyword(name);
	if ( anInt > 0 && yylval.intval == -2 ) {

		switch ( anInt ) {
			case TOKEN_DEFINE:
			    BEGIN START_DEF;
				break;

		    case TOKEN_REPEAT:
			    BEGIN START_TEXT;
				break;

		    case TOKEN_MACRO:
			    BEGIN START_MACRO;
				break;

			case TOKEN_IF:
				BEGIN START_IF;
				break;

			case TOKEN_ELSE:
				BEGIN START_IF;
				break;
		};

		return anInt;
	}

	/* Format specific directive */
	yylval.strval= LIB_strdup( &name[1] );
	return TOKEN_DIRECTIVE;
}


/*
	Function......: do_IDE
	Description...:
	Parameters....: None. */
static int do_IDE(char *name)
{
	register int anInt;

	/* See if this is the BASE token */
	if ( !strcmp("$$", name) ) return TOKEN_BASE;

	/* See if is a reserved keyword */
	if ( (anInt= IsKeyword(name)) == TOKEN_INSN ) {
		/* If it is an instruction ... */
		if ( !Instruction ) {
			/* ... return it if we haven't read one yet. */
			Instruction= TRUE;
			return anInt;
		};

		/* Otherwise it is an instruction used as an operator. */
		switch ( yylval.intval ) {
			case I_AND: return TOKEN_AND;
			case I_OR : return TOKEN_OR;
			case I_NOT: return TOKEN_NOT;
			case I_XOR: return TOKEN_XOR;
			case I_SHR: return TOKEN_SHR;
			case I_SHL: return TOKEN_SHL;
			case I_DIV: return TOKEN_DIV;
		};
	} else if ( anInt == TOKEN_DEF ) {
		old_state= YYSTATE;
		BEGIN READ_ID;
	} else if ( anInt == TOKEN_DDATA ) {
		Instruction= TRUE;
	}

	if ( anInt >= 0 ) return anInt;

	/* Ok, the lexeme is an identifier, look for it in
	   the symbol table */
	yylval.identifier= LookUp(name, IdCase);

	/* If this is a macro local label, link it to the CurrMacroSym */
	if ( !strncmp("%%", name, 2) && yylval.identifier->Head == NULL ) {
		LinkSymbol(GetMacroSym(), yylval.identifier);
	} else if ( yylval.identifier->Type == NO_TYPE ) {
		if ( YYSTATE == START_DEF ) {
			BEGIN READ_DEF;
		} else if ( YYSTATE == START_MACRO ) {
		/* Start state for the reading of macro parameters */
			BEGIN READ_PARAM;
		} else if ( YYSTATE == READ_ID ) {
			BEGIN old_state;
		};
	} else if ( YYSTATE != READ_ID && IDType(yylval.identifier) == MACRO ) {
		if ( IDFlag(yylval.identifier) == DEFINED ) {
		    SwitchToMacro( yylval.identifier, NULL );
		} else {
		/* Begin invokation's parameters reading */
			SetMacro= yylval.identifier;
			BEGIN READ_FORMAL;
		};

		return 0;
	} else if ( YYSTATE == READ_ID ) {
		BEGIN old_state;
	};

	return TOKEN_ID;
}


/*
	Function......: read_text
	Description...: Read text from the current buffer until it reaches
					an ending char or EOF.
					Returns the number of lines readed.
					The text readed is left into the yytext global
					variable.
	Parameters....:
					endc		Character that will end the reading.
					InclEndc	If true the ending character is included
								in the collected text.

					Expand		If true macro parameters are expanded. */
static int read_text(int endc, int InclEndc, int Expand)
{
	register int c;
	register lines= 0;
	unsigned oldFlags;

	oldFlags= ( Expand ?
		SetScanFlag( EXPAND | IGNORE ) :
		UnsetScanFlag(EXPAND) );

	/* Store text in 'yytext'. */
	yylen= 0;
	do {
		c= advance();
		if ( c == '\n') lines++;
	} while ( c != endc && c != EOF_CHAR );

	/* Unexpected end of file. */
	if ( c == EOF_CHAR ) error(ERR_FATAL, ENDERR);

	/* Restore old flags. */
	UnsetScanFlag(0x0ffffffff);
	SetScanFlag(oldFlags);

	if ( !InclEndc ) yylen--;
	yytext[yylen]= '\0';
	return lines;
}


/*
	Function......: AddToHeap
	Description...: Adds a string to the dinamic array LineHeap.
					Returns the length of the string that has been
					added.

					NOTE: The function makes some changes to the
						  passed string, make a copy of it if you
						  need the string after the function call.
	Parameters....:
					text		pointer to the text to be added. */
static int AddToHeap(char *text, int (*stop_func)(char *))
{
	register int textlen= strlen(text);
	register char *crsr;
	int temp;

	/* Nothing to add. */
	if ( textlen == 0 ) return 0;

	/* Skip blanks */
	while ( *text && (*text == ' ' || *text == '\t' || *text == '\b') ) {
		text++;
	};

	/* Nothing to add. */
	if ( !(textlen= strlen(text)) ) return 0;

	if ( text[textlen - 1] == '\n') {
		HeapLines++;
	};

	/* Look for the ending delimiter */
	if ( text[0] == '.' && idechar(text[1]) ) {
		crsr= &text[1];

		/* Isolates the identifier. */
		while ( *crsr && (idechar(*crsr) || isdigit(*crsr)) ) crsr++;
		temp = *crsr;
		*crsr= '\0';

		if ( !stop_func(text) ) {
			crsr= text + strlen(text) - 1;
			/* Push back the delimiter into the scanning buffer. */
			unput(temp);
			while ( crsr >= text ) unput(*crsr--);
			return 0;
		}

		/* Not found! */
		*crsr = temp;
	}

	/* Recalculate the length of the text to be added. */
	textlen= strlen(text);

	/* Allocates room for the new text. */
	if ( (textlen + 1) > (HeapPtr - HeapSize) ) {
		LineHeap= (char *)LIB_realloc(LineHeap, HeapSize + 2048);
	};

	strcpy(&LineHeap[HeapPtr], text);
	HeapPtr += textlen;
	return textlen;
}


/*
	Function......: macro_body_stop
	Description...:
	Parameters....: None. */
static int macro_body_stop(char *text)
{
	switch ( IsKeyword(text) ) {
		case TOKEN_REPEAT:
		case TOKEN_MACRO:
		case TOKEN_IF:
			DefineDepth++;
			break;

		case TOKEN_END:
			/* We have found the .END delimiter.
			   If this encloses all the macro body, do not add
			   the line and return 0. */
			if ( !DefineDepth-- ) return 0;
	};

	return -1;
}


/*
	Function......: if_body_stop
	Description...:
	Parameters....: None. */
static int if_body_stop(char *text)
{
	int anInt =IsKeyword(text);
	switch ( anInt ) {
		case TOKEN_REPEAT:
		case TOKEN_MACRO:
		case TOKEN_IF:
			DefineDepth++;
			break;

		case TOKEN_END:
		case TOKEN_ELSE:
			/* We have found the .END or .ELSE delimiter.
			   If this encloses all the if body, do not add
			   the line and return 0. */
			if ( !DefineDepth-- ) return 0;
	};

	return -1;
}

/* -------- End Of File ----------------------------------------------- */
